// Source code for BlasterWorks

// Use for the base station
// Written by Max Rothstein and Richard Speranza
// for ECE 4760 at Cornell University
// Date: 12-16-2013

// --- Includes and Defines ---------------------------------------
#include "trtSettings.h"
#include "trtkernel644.c"
#include <stdio.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>

// serial communication library
// Don't mess with the semaphores
#define SEM_RX_ISR_SIGNAL 1
#define SEM_STRING_DONE 2 // user hit <enter>
#include "trtUart.h"
#include "trtUart.c"
// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);
#include "wireless.c"
#include "wireless.h"
#define PIPE_ADDRESS_0 0xb4b5b600

// semaphore to protect shared variable
#define SEM_SHARED 3

#define F_CPU 16000000UL       // Set CPU clock to 16 MHz
// default game values
#define MAX_AMMO        10
#define MAX_HEALTH      100
#define NUM_PLAYERS     2

// wireless state constants
#define INIT_RX     0
#define POLL_UPDATE 1
#define INIT_TX     2
#define SEND_UPDATE 3
#define START_GAME  4
#define SEND_GAME_INFO  5

// lcd state constants
#define SELF_INFO   0
#define LEADERBOARD 1

// blaster types
#define SINGLE_SHOT	    0
#define AUTOMATIC	    1

// --- Function Prototypes -----------------------------------
void initialize(void);
void GRN_toggle(void);
void RED_toggle(void);
void task_wireless(void*);
int startup(int);

// --- Global Variables --------------------------------------
int ammo[NUM_PLAYERS] = {MAX_AMMO,MAX_AMMO};
int health[NUM_PLAYERS] = {MAX_HEALTH,MAX_HEALTH};
int args[2] ;

char init_game = 1;
char accepting_input = 1;
int startup_vals[4];  // dummy variables

char data[] = "          ";

// ---- Tasks ------------------------------------------
void task_wireless(void* args) {
    long release, deadline;
	PAYLOAD_LENGTH = 32; //set the payload length of each packet
    char w_state = START_GAME;
	char command; // either S for startup or U for update
    
    // update variables
    int player_num, health_update, ammo_update;
    
    // timeout variables
    char first_entry = 1;
    long enter_time = trtCurrentTime();
    
	fprintf(stdout,"starting base station\n");
	init_RF(0); //disable enhanced shockburst
	char* packet = (char *) malloc(PAYLOAD_LENGTH);
	set_mode(RX_MODE);
    set_RX_addr(PIPE_ADDRESS_0, 0);
    initialize();
	
	while (1) {
        // enters start game on startup and when another game ends
        if ( w_state == START_GAME ) {
            w_state = INIT_RX;
            fprintf(stdout,"WELCOME TO THE LAZER ZONE.\nPrepare...FOR BATTLE!\n");
            int startup_state = 1;
            accepting_input = 1;
            // Prompt user to enter configurable preferences
            while ( accepting_input == 1 ) {
                startup_state = startup(startup_state);
            }
        
            // initialize TX mode
            flush_RX_FIFO();
            set_mode(TX_MODE);
            set_TX_addr(PIPE_ADDRESS_0);
            
            // startup vals are set in startup()
            sprintf(data,"S %d %d %d",startup_vals[0],startup_vals[1],startup_vals[2]);
            
            // initialize health and ammo with desired values
            for ( int i = 0; i < NUM_PLAYERS; i++ ) {
                health[i] = startup_vals[0];
                ammo[i] = startup_vals[1];
            }
            strncpy(packet, data, PAYLOAD_LENGTH); // copy data to sized packet
            first_entry = 1; // used for timeout
            w_state = SEND_GAME_INFO;
        }
        else if ( w_state == SEND_GAME_INFO ) {
            RED_toggle();
            transmit_packet(packet);
            // set entry time on first entry
             if(first_entry == 1) {
                first_entry = 0;
                enter_time = trtCurrentTime();
            }
            // send packet as many times as possible within 5 seconds
            if( (trtCurrentTime() - enter_time) > SECONDS2TICKS(5)) {
                w_state = INIT_RX;
            }
        }
        // initialize transceiver as a areciever
        else if ( w_state == INIT_RX ) {
            flush_RX_FIFO();
            set_mode(RX_MODE); // set to receive mode 
            w_state = POLL_UPDATE;
            first_entry = 1; // used for timeout
        }
        // poll for game state updates
        else if ( w_state == POLL_UPDATE ) {
            if (RX_FIFO_status() == 1) {
                read_RX_payload(packet);

                // copy udpate values to temp variables
                int ammo_update[2];
                int health_update[2];
                sscanf(packet,"%c %d %d %d %d",&command,&health_update[0],&ammo_update[0],
                            &health_update[1],&ammo_update[1]);
                            
                fprintf(stdout,"Update!\n");
                // only process packets labeled with a 'U'
                if ( command == 'U' ) {
                    for ( int i = 0 ; i < NUM_PLAYERS ; i++ ) {
                        // only accept smaller health or ammo values as updates, then
                        // save to global variables
                        if ( health_update[i] < health[i] ) {
                            health[i] = health_update[i];
                        }
                        if ( ammo_update[i] != -1 && ammo_update[i] < ammo[i] ) {
                            ammo[i] = ammo_update[i];
                        }
                        // print updated scoreboard
                        fprintf(stdout,"Player %d: \nHealth: %d  Ammo (pistol): %d\n",i+1,health[i],ammo[i]);
                    }
                    // relay all updates to the blasters
                    w_state = INIT_TX;
                }
            }
        }
        // initialize the transmitter
        else if ( w_state == INIT_TX ) {
            flush_RX_FIFO();
            set_mode(TX_MODE);
            set_TX_addr(PIPE_ADDRESS_0);
            w_state = SEND_UPDATE;
        }
        else if ( w_state == SEND_UPDATE ) {
            GRN_toggle();
            // capture time when first entering the state
            if(first_entry == 1) {
                first_entry = 0;
                enter_time = trtCurrentTime();
            }
            // properly format packet as an update
            sprintf(data,"U %d %d %d %d",health[0],ammo[0],health[1],ammo[1]);
            strncpy(packet, data, PAYLOAD_LENGTH);
            transmit_packet(packet);
            
            // timeout after 1 second
            if( (trtCurrentTime() - enter_time) > SECONDS2TICKS(1)) {
                w_state = INIT_RX;
                for ( int i = 0; i < NUM_PLAYERS; i++ ) {
                    // if any player reaches 0 or below, the game ends since 
                    // there are only two players, return to startup state
                    // and await reinitialization
                    if ( health[i] <= 0 ) {
                        fprintf(stdout,"Player %d has been defeated.\n",i+1);
                        fprintf(stdout,"Play again!\n");
                        w_state = START_GAME;
                        break;
                    }
                }
            }
        }
        
        // shorter sleep duration while attempting to send update
        if ( w_state == SEND_UPDATE ) {
            release 	= trtCurrentTime() + SECONDS2TICKS(.005);
            deadline 	= release;
        }
        else { 
            release 	= trtCurrentTime() + SECONDS2TICKS(.100);
            deadline 	= release + SECONDS2TICKS(.050);     
        }
        trtSleepUntil(release,deadline); 
	}
}

// initiates start up sequence, asks user for input on game settings
// sets all values in startup_vals
int startup ( int stage ) {
    char response; // capture character entries
    int val; // capture integer entries
    
    if ( stage == 1 ) {
        fprintf(stdout,"Would you like to play in default mode? (Y/N)\n>");
        fscanf(stdin, "%c", &response);
        if ( response == 'Y' || response == 'y' ) {
            fprintf(stdout,"Okay, let the games begin.\n");
            startup_vals[0] = 100;
            startup_vals[1] = 10;
            startup_vals[2] = 100;
            accepting_input = 0;
        }
        else if ( response == 'N' || response == 'n' ) {
            fprintf(stdout,"Okay, let's set some preferences.\n");
            stage += 1;
        }
        else {
            fprintf(stdout,"Invalid response, let's try again.\n");
        }
    }
    else if ( stage == 2 ) {
        fprintf(stdout,"Max health: ");
        fscanf(stdin, "%d", &val);
        if ( val <= 0 ) {
            fprintf(stdout,"Invalid response, please enter a number greater than zero.\n");
        }
        else {
            startup_vals[0] = val;
            stage += 1;
        }
    }
    else if ( stage == 3 ) {
        fprintf(stdout,"Max ammo (pistol): ");
        fscanf(stdin,"%d",&val);
        if ( val <= 0 ) {
            fprintf(stdout,"Invalid response, please enter a number greater than zero.\n");
        }
        else {
            startup_vals[1] = val;
            stage += 1;
        }
    }
    else if ( stage == 4 ) {
        fprintf(stdout,"Max ammo (auto): ");
        fscanf(stdin,"%d",&val);
        if ( val <= 0 ) {
            fprintf(stdout,"Invalid response, please enter a number greater than zero.\n");
        }
        else {
            startup_vals[2] = val;
            stage += 1;
            accepting_input = 0;
        }
    }
    return stage;
}

// --- Main Program ----------------------------------
int main(void) {
	//init the UART
	trt_uart_init();
	stdout = stdin = stderr = &uart_str;
	fprintf(stdout,"\n\r TRT Initialized\n\r\n\r");

	// start TRT
	trtInitKernel(80); // 80 bytes for the idle task stack

	// --- create semaphores ----------
	// You must creat the first two semaphores if you use the uart
	trtCreateSemaphore(SEM_RX_ISR_SIGNAL, 0) ; // uart receive ISR semaphore
	trtCreateSemaphore(SEM_STRING_DONE,0) ;  // user typed <enter>

	// variable protection
	trtCreateSemaphore(SEM_SHARED, 1) ; // protect shared variables

	// --- creat tasks  ----------------
	trtCreateTask(task_wireless, 1000, SECONDS2TICKS(0.1), SECONDS2TICKS(0.1), &(args[0]));

	// --- Idle task --------------------------------------
	while (1);

} // main

// --- Other Functions ---------------------------------------
void initialize(void) {
    // LED and button initialization
    PORTC = 0;
    DDRC = ( 1 << PINC2 ) | ( 1 << PINC3 )  ;
}
// toggle the red LED
void RED_toggle(void) {
	PORTC = PORTC ^ 0x04;	
}
// toggle the green LED
void GRN_toggle(void) {
	PORTC = PORTC ^ 0x08; 
}